## ---------
import math, re
import numpy as np


### PART 1: Build the core semantic space

from composes.utils import io_utils
from composes.semantic_space.space import Space

# Build core semantic space from cooccurrnce counts in sparse format
core_space = Space.build( data = "in/core.sm", 
						rows = "in/core.rows",
						cols = "in/core.cols",
						format = "sm")

#print the elements of the space
print core_space.id2row[0:10]
#print the co-occurrence matrix of the space
print core_space.cooccurrence_matrix

# Export space in dense format
core_space.export("out/CORE_SS", format = "dm")
# Or save the Space object in pickle format
io_utils.save(core_space, "out/CORE_SS.pkl")


## Paramaterize the space

# Apply pPMI transformation to core space
from composes.transformation.scaling.ppmi_weighting import PpmiWeighting
core_space = core_space.apply(PpmiWeighting())

# Reduce the 10K dimensions to 50 with Non-negative Matrix Factorization (NMF)
from composes.transformation.dim_reduction.nmf import Nmf
core_space = core_space.apply(Nmf(50))

# Apply row normalization to space
from composes.transformation.scaling.row_normalization import RowNormalization
core_space = core_space.apply(RowNormalization())

# Save the transformed core space 
io_utils.save(core_space, 'out/CORE_SS.ppmi.nmf_50.norm.pkl')


## ---------

### PART 1b: Evaluation of Constituent vectors
###			Correlation with MEN similarity scores

from composes.utils import scoring_utils
from composes.similarity.cos import CosSimilarity

phrase_pairs = io_utils.read_tuple_list('in/MEN_a-and-n_test.txt', fields=[0,1])
gold = io_utils.read_list('in/MEN_a-and-n_test.txt', field=2)

predicted = core_space.get_sims(phrase_pairs, CosSimilarity())

print "Spearman Correlation:", scoring_utils.score(gold, predicted, "spearman")


## ---------

### PART 2: Build peripheral (AN) space based on same contexts + parameters as core space

from composes.semantic_space.peripheral_space import PeripheralSpace

per_space = PeripheralSpace.build( core_space,
									data = "in/an.sm", 
									cols = "in/core.cols",
									format="sm")

#print the elements of the space
print per_space.id2row
#print the co-occurrence matrix of the space
print per_space.cooccurrence_matrix

# Save the peripheral space
io_utils.save(per_space, 'out/PER_SS.ppmi.nmf_50.norm.pkl')


## ---------

### PART 3: Composition models 

train_data = io_utils.read_tuple_list('in/train.txt', fields=[0,1,2])
test_data = io_utils.read_tuple_list('in/test.txt', fields=[0,1,2])


# ADDITION MODELS
from composes.composition.weighted_additive import WeightedAdditive

# Simple additive
add_model = WeightedAdditive(alpha = 1, beta = 1)
add_space = add_model.compose(test_data, core_space)


# Weighted additive (+training)

#train alpha and beta on the training data
wadd_model = WeightedAdditive()
wadd_model.train(train_data, core_space, per_space)

#you can also save the comp function in pickle format
io_utils.save(wadd_model, "out/wadd_model.pkl")

#print model parameters
print "alpha:", wadd_model.alpha
print "beta:", wadd_model.beta

#you can save model parameters
wadd_model.export("out/wadd_model.params")

#compose test ANs based on learned parameters
wadd_space = wadd_model.compose(test_data, core_space)



# MULTIPLICATION MODEL
from composes.composition.multiplicative import Multiplicative
mult_model = Multiplicative()
mult_space = mult_model.compose(test_data, core_space)



# DILATION MODEL (+training)
from composes.composition.dilation import Dilation

#train a Dilation model on the data
dil_model = Dilation()
dil_model.train(train_data, core_space, per_space)

#print its parameters
print "\nlambda:", dil_model._lambda

#compose test ANs based on learned parameters
dil_space = dil_model.compose(test_data, core_space)



# FULL ADDITIVE (a-la Guevara 2010)
from composes.composition.full_additive import FullAdditive

#train a FullAdditive model on the data
fadd_model = FullAdditive()
fadd_model.train(train_data, core_space, per_space)

#print its parameters
print "\nA:", fadd_model._mat_a_t.transpose()
print "B:", fadd_model._mat_b_t.transpose()

#use the model to compose the train data
fadd_space  = fadd_model.compose(test_data, core_space)



# LEXICAL FUNCTION MODEL (a-la Baroni&Zamparelli 2010)
from composes.composition.lexical_function import LexicalFunction

#train using ridge regression
from composes.utils.regression_learner import RidgeRegressionLearner

rr_learner=RidgeRegressionLearner(param = 2,
                                  intercept = False, 
                                  crossvalidation=False)

lfm_model = LexicalFunction(learner = rr_learner)
print type(lfm_model.regression_learner).__name__

lfm_model.train(train_data, core_space, per_space)

lfm_space  = lfm_model.compose(test_data, core_space)

#print its parameters
print "Lexical function space:" 
print lfm_model.function_space.id2row
cooc_mat = lfm_model.function_space.cooccurrence_matrix
print cooc_mat



## ---------

### PART 4: Evaluation of AN vectors
###			Correlation with Mitchell & Lapata (2010) AN similarity scores

from composes.utils import scoring_utils
from composes.similarity.cos import CosSimilarity

phrase_pairs = io_utils.read_tuple_list('in/ML_ans_gold.txt', fields=[0,1])
gold = io_utils.read_list('in/ML_ans_gold.txt', field=2)

#choose a phrase_space to test (eg, per_space)
predicted = per_space.get_sims(phrase_pairs, CosSimilarity())

print "Spearman Correlation:", scoring_utils.score(gold, predicted, "spearman")




## ---------

### Functions for so-called "plausibility measures" (and more)

# Get distance from observed equivalents
def getCosineObs(obs_space, an_space):
    print "Calculating the distance from the corpus-observed vector..."
    cosineObs = {}
    for p in an_space.id2row:
        cosineObs[p] = an_space.get_sim(p, p, CosSimilarity(), space2=obs_space)
    return(cosineObs)
    
cosineObs = getCosineObs(per_space, lfm_space)


# Get the top (N) nearest neighbors of target elements from a given space
def getNeighbors(target_space, neighbor_space, n):
	print "Extracting nearest neighbors..."
    neighbors = {}
    for p in target_space.id2row:
        neighbors[p] = target_space.get_neighbours(p, n, CosSimilarity(), space2=neighbor_space)
    return(neighbors)
    
neighbors = getNeighbors(lfm_space, core_space, 3)



# Get distance from component noun (of AN bigram)
def getCosineN(core_space, an_space):
    print "Calculating the distance from the component nouns..."
    cosineN = {}
    for p in an_space.id2row:
        cosineN[p] = an_space.get_sim(p, p.split('_')[1], CosSimilarity(), space2=core_space)
    return(cosineN)
    
cosineN = getCosineN(core_space, lfm_space)



# Calculate vector length of target vectors
def getVlength(space):
    print "Computing length of composed vectors..."
    length = {}
    for p in space.id2row:
        length[p] = space.get_row(p).norm()
    return(length)

vlength = getVlength(lfm_space)



# Calculate entropy of the vectors in a space
def getEntropy(space):
	print "Computing entropy of composed vectors..."	
    entropies = {}
    for p in space.id2row:
        currvect = [x if x>0.0 else 0 for x in np.array(space.get_row(p)._mat)[0]]
        entropy = 0.0
        s = np.sum(currvect)
        if s == 0.0:
            entropy = -1000000000
        else:
            entropy = math.log(s)-((1/s)*np.sum([x*math.log(x) if x>0.0 else 0 for x in currvect]))
        entropies[p] = entropy
    return(entropies)

entropy = getEntropy(lfm_space)


# Get "new" denisity measure
def getDensity(neighbors, space):
    print "Computing neighborhood density..."
    density = {}
    for k,v in neighbors.iteritems():
		cosSum = 0.0
        N = 0
        for n in range(len(v)):
        	cosSum +=v[n][1]
            N += 1
            x=1
            while n+x in range(len(v)):
            	cosSum += space.get_sim(v[n][0], v[n+x][0], CosSimilarity())
                N += 1
                x += 1
			density[k] = cosSum/N
    return(density)


# merging two spaces:
obs_space = Space.vstack(core_space, per_space)
density = getDensity(neighbors, obs_space)


def getOldDensity(neighbors):
    print "Computing OLD neighborhood density..."
    density = {}
    for k,v in neighbors.iteritems():
		cosSum = 0.0
        N = 0
        for n in range(len(v)):
        	cosSum +=v[n][1]
            N += 1
		density[k] = cosSum/N
    return(density)


oldDensity = getOldDensity(neighbors)


## ---------

### Additional queries

# merging two spaces:
obs_space = Space.vstack(core_space, per_space)



# getting a subspace based on a list of elements:
sub_elements = ['dog-n', 'car-n']
sub_space = Space(core_space.get_rows(sub_elements), sub_elements, core_space.id2column)
# or
sub_elements = [x for x in obs_space.id2row if not re.search('\_',x) ] #should be == core_space
sub_space = Space(obs_space.get_rows(sub_elements), sub_elements, obs_space.id2column)



# getting individual similarities of target items:
print core_space.get_sim('happy-j', 'good-j', CosSimilarity())
print core_space.get_sim('beach-n', 'bacon-n', CosSimilarity())

print per_space.get_sim('cold-j_air-n', 'hot-j_air-n', CosSimilarity())
print per_space.get_sim('cold-j_air-n', 'hot-j_water-n', CosSimilarity())
